// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.PixelSizes;
import imagescience.images.ShortImage;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;
import imagescience.utilities.Timer;

/** Provides methods for computing the Laplacian of images. */
public final class Laplacian {

    /** Method for computing the Laplacian of images of type
	<code>ByteImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage run(final ByteImage image,
			  final float scale,
			  final boolean zerocross) {

	return run(new FloatImage(image),scale,zerocross,true);
    }
	
    /** Method for computing the Laplacian of images of type
	<code>ShortImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage run(final ShortImage image,
			  final float scale,
			  final boolean zerocross) {

	return run(new FloatImage(image),scale,zerocross,true);
    }
	
    /** Method for computing the Laplacian of images of type
	<code>FloatImage</code>.

	@param image the input image for which the Laplacian needs to be
	computed.<br><ul>If the size of the image in the z-dimension
	equals one, this method will compute, for every pixel, the
	two-dimensional (2D) Laplacian. Otherwise it will compute for
	every pixel the full three-dimensional (3D) Laplacian. These
	computations are applied to every volume in the fourth
	dimension.</ul>

	@param scale the smoothing scale at which the required image
	derivatives are computed.<br><ul>The scale is equal to the
	standard deviation of the Gaussian kernel used for
	differentiation. In order to enforce physical isotropy, for each
	dimension, the scale is divided by the pixel size of the image
	in that dimension. If the scale parameter is negative, the
	Laplacian computation is skipped. This allows the user to apply
	zero-crossing detection independently to an image.</ul>

	@param zerocross determines whether or not zero-crossings are
	detected.<br><ul>To determine whether or not there is a
	zero-crossing, this method compares the signs of neighboring
	pixels, and if they are different, it uses a linear interpolation
	scheme to determine the pixel whose center is closest to the
	location of the crossing. This pixel is then set to 255. Other
	pixels are put to zero.</ul>

	@param useimage indicates whether or not the contents of the
	input image may be altered by the method.<br><ul>If this
	parameter is <code>true</code>, the method may use the input
	image rather than a newly created image to store intermediate
	results, thereby saving memory. If it is <code>false</code>, the
	input image is left unaltered.</ul>
	
	@return an image showing the (zero-crossings of the) Laplacian
	of the input image. If none of the operations of Laplacian
	computation or zero-crossing detection is carried out, the
	method returns a handle to (a copy of) the input image.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage run(final FloatImage image,
			  final float scale,
			  final boolean zerocross,
			  final boolean useimage) {
	
	Features.printVersion("Laplacian");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	msg.debug("Checking arguments for Laplacian computation");
	boolean bLaplacian = true;
	if (scale < 0.0f) bLaplacian = false;

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	msg.debug("Pixel size in x-dimension: " + ps.x);
	msg.debug("Pixel size in y-dimension: " + ps.y);
	msg.debug("Pixel size in z-dimension: " + ps.z);

	FloatImage lapImage = useimage ? image : image.duplicate();

	final boolean bDebug = msg.debug();

	if (dims.z == 1) { // 2D case

	    // Compute Laplacian if requested:
	    if (bLaplacian) {
		FloatImage Ixx, Iyy;
		if (Features.sepSmooth()) {
		    msg.debug("Seperating smoothing and differentiation");
		    msg.debug("Gaussian smoothing at scale "+scale);
		    msg.debug(false);
		    differ.smooth(lapImage,scale,false);
		    msg.debug(bDebug);
		    msg.debug("Computing Ixx");
		    msg.debug(false);
		    Ixx = differ.fdiff(lapImage,0.0f,2,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iyy");
		    msg.debug(false);
		    Iyy = differ.fdiff(lapImage,0.0f,0,2,0,false);
		    msg.debug(bDebug);
		} else {
		    msg.debug("Combining smoothing and differentiation");
		    msg.debug("Gaussian differentiation at scale "+scale);
		    msg.debug("Computing Ixx");
		    msg.debug(false);
		    Ixx = differ.gauss(lapImage,scale,2,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iyy");
		    msg.debug(false);
		    Iyy = differ.gauss(lapImage,scale,0,2,0,false);
		    msg.debug(bDebug);
		}

		// Compute 2D Laplacian image:
		msg.debug("Computing 2D Laplacian image");
		lapImage.add(Ixx);
	    }
	    
	    // Apply zero-crossing detection if requested:
	    if (zerocross) {
		msg.debug("Detecting 2D zero crossings");
		final Progressor pgs = new Progressor("Detecting zero crossings");
		pgs.steps(dims.t*dims.z*dims.y);
		lapImage.axes(Axes.X);
		final Coordinates crds0 = new Coordinates();
		final Coordinates crds1 = new Coordinates();
		float[] lap0 = new float[dims.x];
		float[] lap1 = new float[dims.x];
		float[] ltmp = null;
		float[] zer0 = new float[dims.x];
		float[] zer1 = new float[dims.x];
		float[] ztmp = null;
		final int dimsxm1 = dims.x - 1;
		final int dimsym1 = dims.y - 1;
		
		for (crds0.t=crds1.t=0; crds0.t<dims.t; ++crds0.t, ++crds1.t) {
		    crds1.y=0; lapImage.get(crds1,lap1);
		    for (int x=0; x<dims.x; ++x) zer1[x]=fOFF;
		    // Top and intermediate rows:
		    for (crds0.y=0, crds1.y=1; crds0.y<dimsym1; ++crds0.y, ++crds1.y) {
			ltmp=lap0; lap0=lap1; lap1=ltmp; lapImage.get(crds1,lap1);
			ztmp=zer0; zer0=zer1; zer1=ztmp; for (int x=0; x<dims.x; ++x) zer1[x]=fOFF;
			for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1) {
			    // Horizontal zero or sign changes:
			    if (lap0[x]*lap0[xp1] <= 0.0f && (lap0[x]-lap0[xp1]) != 0.0f)
				if (lap0[x]/(lap0[x]-lap0[xp1]) < 0.5f) zer0[x] = fON;
				else zer0[xp1] = fON;
			    // Vertical zero or sign changes:
			    if (lap0[x]*lap1[x] <= 0.0f && (lap0[x]-lap1[x]) != 0.0f)
				if (lap0[x]/(lap0[x]-lap1[x]) < 0.5f) zer0[x] = fON;
				else zer1[x] = fON;
			}
			// Last column vertical zero or sign changes:
			if (lap0[dimsxm1]*lap1[dimsxm1] <= 0.0f && (lap0[dimsxm1]-lap1[dimsxm1]) != 0.0f)
			    if (lap0[dimsxm1]/(lap0[dimsxm1]-lap1[dimsxm1]) < 0.5f) zer0[dimsxm1] = fON;
			    else zer1[dimsxm1] = fON;
			lapImage.set(crds0,zer0);
			pgs.step();
		    }
		    // Last row horizontal zero or sign changes:
		    for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1)
			if (lap1[x]*lap1[xp1] <= 0.0f && (lap1[x]-lap1[xp1]) != 0.0f)
			    if (lap1[x]/(lap1[x]-lap1[xp1]) < 0.5f) zer1[x] = fON;
			    else zer1[xp1] = fON;
		    lapImage.set(crds0,zer1);
		    pgs.step();
		}
		pgs.end();
	    }
	    
	} else { // 3D case

	    // Compute Laplacian if requested:
	    if (bLaplacian) {
		FloatImage Ixx, Iyy, Izz;
		if (Features.sepSmooth()) {
		    msg.debug("Seperating smoothing and differentiation");
		    msg.debug("Gaussian smoothing at scale "+scale);
		    msg.debug(false);
		    differ.smooth(lapImage,scale,false);
		    msg.debug(bDebug);
		    msg.debug("Computing Ixx");
		    msg.debug(false);
		    Ixx = differ.fdiff(lapImage,0.0f,2,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iyy");
		    msg.debug(false);
		    Iyy = differ.fdiff(lapImage,0.0f,0,2,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Izz");
		    msg.debug(false);
		    Izz = differ.fdiff(lapImage,0.0f,0,0,2,false);
		    msg.debug(bDebug);
		} else {
		    msg.debug("Combining smoothing and differentiation");
		    msg.debug("Gaussian differentiation at scale "+scale);
		    msg.debug("Computing Ixx");
		    msg.debug(false);
		    Ixx = differ.gauss(lapImage,scale,2,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iyy");
		    msg.debug(false);
		    Iyy = differ.gauss(lapImage,scale,0,2,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Izz");
		    msg.debug(false);
		    Izz = differ.gauss(lapImage,scale,0,0,2,false);
		    msg.debug(bDebug);
		}

		// Compute 3D Laplacian image:
		msg.debug("Computing 3D Laplacian");
		lapImage.add(Iyy);
		lapImage.add(Ixx);
	    }
	    
	    // Apply zero-crossing detection if requested:
	    if (zerocross) {
		msg.debug("Detecting 3D zero crossings");
		final Progressor pgs = new Progressor("Detecting zero crossings");
		pgs.steps(dims.t*dims.z*dims.y);
		lapImage.axes(Axes.X+Axes.Y);
		final Coordinates crds0 = new Coordinates();
		final Coordinates crds1 = new Coordinates();
		float[][] lap0 = new float[dims.y][dims.x];
		float[][] lap1 = new float[dims.y][dims.x];
		float[][] zer0 = new float[dims.y][dims.x];
		float[][] zer1 = new float[dims.y][dims.x];
		float[][] ltmp, ztmp;
		float[] lap0y, lap0yp1, lap1y, lap1yp1;
		float[] zer0y, zer0yp1, zer1y, zer1yp1;
		final int dimsxm1 = dims.x - 1;
		final int dimsym1 = dims.y - 1;
		final int dimszm1 = dims.z - 1;
		
		for (crds0.t=crds1.t=0; crds0.t<dims.t; ++crds0.t, ++crds1.t) {
		    crds1.z=0; lapImage.get(crds1,lap1);
		    for (int y=0; y<dims.y; ++y) for (int x=0; x<dims.x; ++x) zer1[y][x]=fOFF;
		    // Front and intermediate slices:
		    for (crds0.z=0, crds1.z=1; crds0.z<dimszm1; ++crds0.z, ++crds1.z) {
			ltmp=lap0; lap0=lap1; lap1=ltmp; lapImage.get(crds1,lap1);
			ztmp=zer0; zer0=zer1; zer1=ztmp;
			for (int y=0; y<dims.y; ++y) for (int x=0; x<dims.x; ++x) zer1[y][x]=fOFF;
			// Top and intermediate rows:
			for (int y=0, yp1=1; y<dimsym1; ++y, ++yp1) {
			    lap0y=lap0[y]; lap0yp1=lap0[yp1]; lap1y=lap1[y];
			    zer0y=zer0[y]; zer0yp1=zer0[yp1]; zer1y=zer1[y];
			    for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1) {
				// Horizontal zero or sign changes:
				if (lap0y[x]*lap0y[xp1] <= 0.0f && (lap0y[x]-lap0y[xp1]) != 0.0f)
				    if (lap0y[x]/(lap0y[x]-lap0y[xp1]) < 0.5f) zer0y[x] = fON;
				    else zer0y[xp1] = fON;
				// Vertical zero or sign changes:
				if (lap0y[x]*lap0yp1[x] <= 0.0f && (lap0y[x]-lap0yp1[x]) != 0.0f)
				    if (lap0y[x]/(lap0y[x]-lap0yp1[x]) < 0.5f) zer0y[x] = fON;
				    else zer0yp1[x] = fON;
				// Depth zero or sign changes:
				if (lap0y[x]*lap1y[x] <= 0.0f && (lap0y[x]-lap1y[x]) != 0.0f)
				    if (lap0y[x]/(lap0y[x]-lap1y[x]) < 0.5f) zer0y[x] = fON;
				    else zer1y[x] = fON;
			    }
			    // Last column vertical zero or sign changes:
			    if (lap0y[dimsxm1]*lap0yp1[dimsxm1] <= 0.0f && (lap0y[dimsxm1]-lap0yp1[dimsxm1]) != 0.0f)
				if (lap0y[dimsxm1]/(lap0y[dimsxm1]-lap0yp1[dimsxm1]) < 0.5f) zer0y[dimsxm1] = fON;
				else zer0yp1[dimsxm1] = fON;
			    // Last column depth zero or sign changes:
			    if (lap0y[dimsxm1]*lap1y[dimsxm1] <= 0.0f && (lap0y[dimsxm1]-lap1y[dimsxm1]) != 0.0f)
				if (lap0y[dimsxm1]/(lap0y[dimsxm1]-lap1y[dimsxm1]) < 0.5f) zer0y[dimsxm1] = fON;
				else zer1y[dimsxm1] = fON;
			    pgs.step();
			}
			lap0y = lap0[dimsym1]; lap1y = lap1[dimsym1];
			zer0y = zer0[dimsym1]; zer1y = zer1[dimsym1];
			// Last row horizontal zero or sign changes:
			for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1)
			    if (lap0y[x]*lap0y[xp1] <= 0.0f && (lap0y[x]-lap0y[xp1]) != 0.0f)
				if (lap0y[x]/(lap0y[x]-lap0y[xp1]) < 0.5f) zer0y[x] = fON;
				else zer0y[xp1] = fON;
			// Last row depth zero or sign changes:
			for (int x=0; x<dims.x; ++x)
			    if (lap0y[x]*lap1y[x] <= 0.0f && (lap0y[x]-lap1y[x]) != 0.0f)
				if (lap0y[x]/(lap0y[x]-lap1y[x]) < 0.5f) zer0y[x] = fON;
				else zer1y[x] = fON;
			lapImage.set(crds0,zer0);
			pgs.step();
		    } // end z-loop
		    // Last slice:
		    for (int y=0, yp1=1; y<dimsym1; ++y, ++yp1) {
			lap1y=lap1[y]; lap1yp1=lap1[yp1];
			zer1y=zer1[y]; zer1yp1=zer1[yp1];
			for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1) {
			    // Horizontal zero or sign changes:
			    if (lap1y[x]*lap1y[xp1] <= 0.0f && (lap1y[x]-lap1y[xp1]) != 0.0f)
				if (lap1y[x]/(lap1y[x]-lap1y[xp1]) < 0.5f) zer1y[x] = fON;
				else zer1y[xp1] = fON;
			    // Vertical zero or sign changes:
			    if (lap1y[x]*lap1yp1[x] <= 0.0f && (lap1y[x]-lap1yp1[x]) != 0.0f)
				if (lap1y[x]/(lap1y[x]-lap1yp1[x]) < 0.5f) zer1y[x] = fON;
				else zer1yp1[x] = fON;
			}
			// Last column vertical zero or sign changes:
			if (lap1y[dimsxm1]*lap1yp1[dimsxm1] <= 0.0f && (lap1y[dimsxm1]-lap1yp1[dimsxm1]) != 0.0f)
			    if (lap1y[dimsxm1]/(lap1y[dimsxm1]-lap1yp1[dimsxm1]) < 0.5f) zer1y[dimsxm1] = fON;
			    else zer1yp1[dimsxm1] = fON;
			pgs.step();
		    }
		    lap1y=lap1[dimsym1]; zer1y=zer1[dimsym1];
		    // Last row horizontal zero or sign changes:
		    for (int x=0, xp1=1; x<dimsxm1; ++x, ++xp1)
			if (lap1y[x]*lap1y[xp1] <= 0.0f && (lap1y[x]-lap1y[xp1]) != 0.0f)
			    if (lap1y[x]/(lap1y[x]-lap1y[xp1]) < 0.5f) zer1y[x] = fON;
			    else zer1y[xp1] = fON;
		    lapImage.set(crds0,zer1);
		    pgs.step();
		} // end t-loop
		pgs.end();
	    }
	}
	
	tmr.stop(msg);
	return lapImage;
    }

    private final Messenger msg = Features.msg;
    private final Differentiator differ = new Differentiator();

    private final static float fON = 255.0f;
    private final static float fOFF = 0.0f;
}
