// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.utilities;

import java.lang.Math;

/** Provides methods for carrying out a number of elementary
    mathematical operations and function evaluations. */
public final class FMath {

    /** Returns the largest integral value that is not greater than the
        given value. If the given value is already equal to a
        mathematical integer, then the result is the same as the given
        value. Note that the method works only for <code>float</code>
        values that fall within the range spanned by the integers. In
        that case, this method gives the same result as the
        corresponding method in Java's <code>Math</code> class, but is
        in general much faster. */
    public static int floor(final float f) {

	if (f >= 0.0f) return (int)f;
	else {
	    final int iAdd = (int)f - 1;
	    return (((int)(f - iAdd)) + iAdd);
	}
    }

    /** Returns the largest integral value that is not greater than the
        given value. If the given value is already equal to a
        mathematical integer, then the result is the same as the given
        value. Note that the method works only for <code>double</code>
        values that fall within the range spanned by the integers. In
        that case, this method gives the same result as the
        corresponding method in Java's <code>Math</code> class, but is
        in general much faster. */
    public static int floor(final double d) {

	if (d >= 0.0) return (int)d;
	else {
	    final int iAdd = (int)d - 1;
	    return (((int)(d - iAdd)) + iAdd);
	}
    }

    /** Returns the smallest integral value that is not less than the
        given value. If the given value is already equal to a
        mathematical integer, then the result is the same as the given
        value. Note that the method works only for <code>float</code>
        values that fall within the range spanned by the integers. In
        that case, this method gives the same result as the
        corresponding method in Java's <code>Math</code> class, but is
        in general much faster. */
    public static int ceil(final float f) {

	final float mf = -f;
	if (mf >= 0.0f) return -((int)mf);
	else {
	    final int iAdd = (int)mf - 1;
	    return -(((int)(mf - iAdd)) + iAdd);
	}
    }

    /** Returns the smallest integral value that is not less than the
        given value. If the given value is already equal to a
        mathematical integer, then the result is the same as the given
        value. Note that the method works only for <code>double</code>
        values that fall within the range spanned by the integers. In
        that case, this method gives the same result as the
        corresponding method in Java's <code>Math</code> class, but is
        in general much faster.  */
    public static int ceil(final double d) {

	final double md = -d;
	if (md >= 0.0) return -((int)md);
	else {
	    final int iAdd = (int)md - 1;
	    return -(((int)(md - iAdd)) + iAdd);
	}
    }

    /** Returns the integral value closest to the given value. Note that
        the method works only for <code>float</code> values that fall
        within the range spanned by the integers. In that case, this
        method gives the same result as the corresponding method in
        Java's <code>Math</code> class, but is in general much faster. */
    public static int round(final float f) {

	final float f05 = f + 0.5f;
	if (f05 >= 0.0) return (int)f05;
	else {
	    final int iAdd = (int)f05 - 1;
	    return (((int)(f05 - iAdd)) + iAdd);
	}
    }

    /** Returns the integral value closest to the given value. Note that
        the method works only for <code>double</code> values that fall
        within the range spanned by the integers. In that case, this
        method gives the same result as the corresponding method in
        Java's <code>Math</code> class, but is in general much faster. */
    public static int round(final double d) {

	final double d05 = d + 0.5;
	if (d05 >= 0.0) return (int)d05;
	else {
	    final int iAdd = (int)d05 - 1;
	    return (((int)(d05 - iAdd)) + iAdd);
	}
    }

    /** Returns the given value clipped to the given range. Guaranteed
        to work properly only if <code>min <= max</code>. */
    public static float clip(final float f, final float min, final float max) {

	return ((f > max) ? max : ((f < min) ? min : f));
    }

    /** Returns the given value clipped to the given range. Guaranteed
        to work properly only if <code>min <= max</code>. */
    public static double clip(final double f, final double min, final double max) {

	return ((f > max) ? max : ((f < min) ? min : f));
    }

    /** Returns the minimum of the two given values. */
    public static float min(final float f1, final float f2) {

	return ((f1 < f2) ? f1 : f2);
    }
    
    /** Returns the minimum of the two given values. */
    public static double min(final double d1, final double d2) {

	return ((d1 < d2) ? d1 : d2);
    }
    
    /** Returns the minimum of the four given values. */
    public static float min(final float f1, final float f2,
			    final float f3, final float f4) {

	final float f12 = min(f1,f2);
	final float f34 = min(f3,f4);
	return ((f12 < f34) ? f12 : f34);
    }
    
    /** Returns the minimum of the four given values. */
    public static double min(final double d1, final double d2,
			     final double d3, final double d4) {

	final double d12 = min(d1,d2);
	final double d34 = min(d3,d4);
	return ((d12 < d34) ? d12 : d34);
    }
    
    /** Returns the minimum of the eight given values. */
    public static float min(final float f1, final float f2,
			    final float f3, final float f4,
			    final float f5, final float f6,
			    final float f7, final float f8) {

	final float f1234 = min(f1,f2,f3,f4);
	final float f5678 = min(f5,f6,f7,f8);
	return ((f1234 < f5678) ? f1234 : f5678);
    }

    /** Returns the minimum of the eight given values. */
    public static double min(final double d1, final double d2,
			     final double d3, final double d4,
			     final double d5, final double d6,
			     final double d7, final double d8) {

	final double d1234 = min(d1,d2,d3,d4);
	final double d5678 = min(d5,d6,d7,d8);
	return ((d1234 < d5678) ? d1234 : d5678);
    }

    /** Returns the maximum of the two given values. */
    public static float max(final float f1, final float f2) {

	return ((f1 > f2) ? f1 : f2);
    }
    
    /** Returns the maximum of the two given values. */
    public static double max(final double d1, final double d2) {

	return ((d1 > d2) ? d1 : d2);
    }
    
    /** Returns the maximum of the four given values. */
    public static float max(final float f1, final float f2,
			    final float f3, final float f4) {

	final float f12 = max(f1,f2);
	final float f34 = max(f3,f4);
	return ((f12 > f34) ? f12 : f34);
    }
    
    /** Returns the maximum of the four given values. */
    public static double max(final double d1, final double d2,
			     final double d3, final double d4) {

	final double d12 = max(d1,d2);
	final double d34 = max(d3,d4);
	return ((d12 > d34) ? d12 : d34);
    }
    
    /** Returns the maximum of the eight given values. */
    public static float max(final float f1, final float f2,
			    final float f3, final float f4,
			    final float f5, final float f6,
			    final float f7, final float f8) {

	final float f1234 = max(f1,f2,f3,f4);
	final float f5678 = max(f5,f6,f7,f8);
	return ((f1234 > f5678) ? f1234 : f5678);
    }

    /** Returns the maximum of the eight given values. */
    public static double max(final double d1, final double d2,
			     final double d3, final double d4,
			     final double d5, final double d6,
			     final double d7, final double d8) {

	final double d1234 = max(d1,d2,d3,d4);
	final double d5678 = max(d5,d6,d7,d8);
	return ((d1234 > d5678) ? d1234 : d5678);
    }

    /** Returns the number of decimals in the given float number. */
    public static int decimals(final float f) {

	if (f == 0.0f) return 1;
	final String s = Float.toString(f);
	// If f is not zero and outside the following range, it is in
	// scientific notation (see Java documentation):
	if ((f > -10000000f && f <= -0.001f) || (f >= 0.001f && f < 10000000f)) {
	    final int iLen = s.length();
	    final int iIndexDot = s.indexOf(".");
	    // For some stupid reason, floats that have "00x" behind the dot are
	    // given an extra zero at the end by Float.toString(). This must be removed:
	    if ((iLen-1-iIndexDot == 4) && s.endsWith("0")) return iLen-iIndexDot-2;
	    else return iLen-iIndexDot-1;
	}
	else {
	    final int iIndexE = s.indexOf("E");
	    final String exp = s.substring(iIndexE + 1);
	    final int iExp = Integer.valueOf(exp).intValue();
	    // Note that there can only be 8 decimals in a Java float:
	    if (iExp >= 7) return 1;
	    final int iIndexDot = s.indexOf(".");
	    final int iDecsBetween = iIndexE - iIndexDot - 1;
	    // If there is just one decimal "0" behind the dot, it can be left out:
	    if (iDecsBetween == 1 && s.substring(iIndexDot+1,iIndexDot+2).equals("0")) return iDecsBetween - iExp - 1;
	    else return iDecsBetween - iExp;
	}
    }

    private static final double[] lngammac = { 76.18009172947146, -86.50532032941677,
					       24.01409824083091, -1.231739572450155,
					       0.1208650973866179E-2, -0.5395239384953E-5 };
    
    /** Returns the natural logarithm of <code>Gamma(x)</code> for
        <code>x > 0</code>. Computed using the approximation described
        by C. Lanczos, "A Precision Approximation of the Gamma Function",
	SIAM Journal on Numerical Analysis, Series B, vol. 1, 1964,
	pp. 86-96. Based on the algorithm given by W. H. Press et al.,
	Numerical Recipes in C: The Art of Scientific Computing, 2nd
	ed., Cambridge University Press, Cambridge, 1992, Section 6.1. */
    public static double lngamma(final double x) {

	double y = x;
	double t = x + 5.5;
	t -= (x + 0.5)*Math.log(t);
	double ser = 1.000000000190015;
	for (int i=0; i<=5; ++i) ser += lngammac[i]/++y;

	return -t + Math.log(2.5066282746310005*ser/x);
    }
    
    private static final double invsqrt2pi = 1.0/Math.sqrt(2*Math.PI);

    /** Returns the value at <code>x</code> of the Gaussian distribution
        with standard deviation <code>sd</code>. The distribution is
        normalized to unit integral. */
    public static double gauss(final double x, final double sd) {

	return invsqrt2pi*Math.exp(-x*x/(2*sd*sd))/sd;
    }
    
    /** Returns the value of the <code>sinc</code> function at
        <code>x</code>. The <code>sinc</code> function is defined as
        <code>sin(x)/x</code>. */
    public static double sinc(final double x) {

	if (x == 0.0) return 1.0;
	else return Math.sin(x)/x;
    }

    /** Returns the value at <code>x</code> of the zeroth-order Bessel
        function of the first kind. Based on the algorithm given by
        W. H. Press et al., Numerical Recipes in C: The Art of
        Scientific Computing, 2nd ed., Cambridge University Press,
        Cambridge, 1992, Section 6.5. */
    public static double besselj0(final double x) {

	double ax, z, xx, y, a1, a2;

	if ((ax = Math.abs(x)) < 8) { // Direct rational approximation.
	    y = x*x;
	    a1 = 57568490574.0+y*(-13362590354.0+y*(651619640.7+y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
	    a2 = 57568490411.0+y*(1029532985.0+y*(9494680.718+y*(59272.64853+y*(267.8532712+y*1.0))));
	    return a1/a2;
	} else { // Fitting function.
	    z = 8/ax; y = z*z; xx = ax-0.785398164;
	    a1 = 1+y*(-0.1098628627e-2+y*(0.2734510407e-4+y*(-0.2073370639e-5+y*0.2093887211e-6)));
	    a2 = -0.1562499995e-1+y*(0.1430488765e-3+y*(-0.6911147651e-5+y*(0.7621095161e-6-y*0.934945152e-7)));
	    return Math.sqrt(0.636619772/ax)*(Math.cos(xx)*a1-z*Math.sin(xx)*a2);
	}
    }

    /** Returns the value at <code>x</code> of the first-order Bessel
        function of the first kind. Based on the algorithm given by
        W. H. Press et al., Numerical Recipes in C: The Art of
        Scientific Computing, 2nd ed., Cambridge University Press,
        Cambridge, 1992, Section 6.5. */
    public static double besselj1(final double x) {

	double ax, z, xx, y, a1, a2;

	if ((ax = Math.abs(x)) < 8) { // Direct rational approximation.
	    y = x*x;
	    a1 = x*(72362614232.0+y*(-7895059235.0+y*(242396853.1+y*(-2972611.439+y*(15704.48260+y*(-30.16036606))))));
	    a2 = 144725228442.0+y*(2300535178.0+y*(18583304.74+y*(99447.43394+y*(376.9991397+y*1.0))));
	    return a1/a2;
	} else { // Fitting function.
	    z = 8/ax; y = z*z; xx = ax-2.356194491;
	    a1 = 1+y*(0.183105e-2+y*(-0.3516396496e-4+y*(0.2457520174e-5+y*(-0.240337019e-6))));
	    a2 = 0.04687499995+y*(-0.2002690873e-3+y*(0.8449199096e-5+y*(-0.88228987e-6+y*0.105787412e-6)));
	    a1 = Math.sqrt(0.636619772/ax)*(Math.cos(xx)*a1-z*Math.sin(xx)*a2);
	    if (x < 0) return -a1; else return a1;
	}
    }

    private static final double ACC = 40.0;
    private static final double BIGNO = 1.0e10;
    private static final double BIGNI = 1.0e-10;

    /** Returns the value at <code>x</code> of the Bessel function of
	the first kind of order <code>n >= 0</code>. Based on the
	algorithm given by W. H. Press et al., Numerical Recipes in C:
	The Art of Scientific Computing, 2nd ed., Cambridge University
	Press, Cambridge, 1992, Section 6.5. */
    public static double besselj(final int n, final double x) {

	// Order 0 or 1:
	if (n == 0) return besselj0(x);
	if (n == 1) return besselj1(x);

	// Order 2 and larger:
	if (x == 0) return 0;
	final double ax = Math.abs(x);
	double bjm, bjp, sum, ans;
	if (ax > n) { // Upward recurrence from J0 and J1.
	    final double tox = 2/ax;
	    bjm = besselj0(ax);
	    ans = besselj1(ax);
	    for (int j=1; j<n; ++j) {
		bjp = j*tox*ans - bjm;
		bjm = ans;
		ans = bjp;
	    }
	} else { // Downward recurrence from an even m.
	    final double tox = 2/ax;
	    final int m = 2*((n + (int)Math.sqrt(ACC*n))/2);
	    boolean jsum = false;
	    bjp = ans = sum = 0;
	    double bj = 1;
	    for (int j=m; j>0; --j) {
		bjm = j*tox*bj - bjp;
		bjp = bj;
		bj = bjm;
		if (Math.abs(bj) > BIGNO) {
		    bj *= BIGNI;
		    bjp *= BIGNI;
		    ans *= BIGNI;
		    sum *= BIGNI;
		}
		if (jsum) sum += bj;
		jsum = !jsum;
		if (j == n) ans = bjp;
	    }
	    sum = 2*sum - bj;
	    ans /= sum;
	}
	return (x < 0 && (n & 1) == 1) ? -ans : ans;
    }

    /** Returns the value at <code>(u,v)</code> of the series expansion
        of the Lommel U function of order <code>n >= 0</code> evaluated
        to <code>t > 0</code> terms. Straightforward implementation of
        the formula given by M. Born and E. Wolf, Principles of Optics,
        5th ed., Pergamon Press, Oxford, 1975, Section 8.8. */
    public static double lommelu(final int n, final int t, final double u, final double v) {

	final double uov2 = u*u/(v*v);
	double uovn = Math.pow(u/v,n);
	double sum = 0; int sign = 1;
	for (int s=0; s<t; ++s) {
	    sum += sign*uovn*besselj(n+2*s,v);
	    uovn *= uov2; sign *= -1;
	}
	return sum;
    }

    /** Returns the value at <code>(u,v)</code> of the series expansion
        of the Lommel V function of order <code>n >= 0</code> evaluated
        to <code>t > 0</code> terms. Straightforward implementation of
        the formula given by M. Born and E. Wolf, Principles of Optics,
        5th ed., Pergamon Press, Oxford, 1975, Section 8.8. */
    public static double lommelv(final int n, final int t, final double u, final double v) {

	final double vou2 = v*v/(u*u);
	double voun = Math.pow(v/u,n);
	double sum = 0; int sign = 1;
	for (int s=0; s<t; ++s) {
	    sum += sign*voun*besselj(n+2*s,v);
	    voun *= vou2; sign *= -1;
	}
	return sum;
    }

    private static final int MAXBIT = 30;

    /** Indicates whether the given integer is a power of 2. Returns
	<code>true</code> if it is and <code>false</code> if it is not a
	power of 2. Works for integers upto and including 2^30. */
    public static boolean power2(final int i) {

	for (int p=0, p2=1; p<=MAXBIT; ++p, p2<<=1)
	    if (i == p2) return true;
	return false;
    }

}
