//**********************************************************************
//COPYRIGHT NOTICE AND DISCLAIMER
//
//Copyright (c) Erik Meijering. All rights reserved.
//
//Permission to copy and use this software and accompanying
//documentation for educational, research, and not-for-profit
//purposes, without fee and without a signed licensing agreement,
//is hereby granted, provided that the above copyright notice, this
//paragraph and the following two paragraphs appear in all copies.
//The copyright holder is free to make upgraded or improved versions
//of the software available for a fee or commercially only.
//
//IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
//DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
//AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
//POSSIBILITY OF SUCH DAMAGE.
//
//THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
//INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
//SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
//THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
//SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
//**********************************************************************
package neuronJgui;

import ij.IJ;
import ij.gui.GUI;
import ij.text.TextPanel;
import ij.text.TextWindow;
import imagescience.utilities.FMath;

import java.awt.Button;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Dialog;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Label;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

public final class MeasurementsDialog extends Dialog implements ActionListener{

    private final Choice typeChoice;
    private final Choice clusterChoice;

    private final Checkbox statistCheckbox;
    private final Checkbox individCheckbox;
    private final Checkbox clearCheckbox;

    private final Button runButton;
    private final Button closeButton;

    private final String statistHeadings = "Image\tCluster\tType\tCount\tSum\tMean\tStDev\tMin\tMax";
    private final String individHeadings = "Image\tTracing\tCluster\tType\tLabel\tLength";

    private final GridBagConstraints c = new GridBagConstraints();
    private final GridBagLayout grid = new GridBagLayout();

    private boolean bFirstChoice = true;
    private boolean bFirstReset = true;

    // Builds the dialog for doing length measurements.
    MeasurementsDialog() {

	super(IJ.getInstance(),"NeuronJ: Measurements",false);
	setLayout(grid);

	// Add choices:
        typeChoice = addChoice("Tracing type:");
        clusterChoice = addChoice("cluster:");

	// Add check boxes:
        c.insets = new Insets(15, 18, 0, 18);
        c.gridx = 0; c.gridy++; c.gridwidth = 3;
        c.anchor = GridBagConstraints.WEST;
        individCheckbox = new Checkbox(" Display individual tracing lengths");
        grid.setConstraints(individCheckbox, c);
        individCheckbox.setState(true);
        add(individCheckbox);

	c.gridy++; c.insets = new Insets(0, 18, 0, 18);
        statistCheckbox = new Checkbox(" Display tracing length statistics");
        grid.setConstraints(statistCheckbox, c);
        statistCheckbox.setState(true);
        add(statistCheckbox);

	c.gridy++;
        clearCheckbox = new Checkbox(" Clear previous measurements");
        grid.setConstraints(clearCheckbox, c);
        clearCheckbox.setState(true);
        add(clearCheckbox);

	// Add Run and Close buttons:
	final Panel buttons = new Panel();
        buttons.setLayout(new FlowLayout(FlowLayout.CENTER, 5, 0));
        runButton = new Button("  Run  ");
        runButton.addActionListener(this);
        closeButton = new Button("Close");
        closeButton.addActionListener(this);
        buttons.add(runButton);
        buttons.add(closeButton);
	c.gridx = 0; c.gridy++; c.gridwidth = 3;
        c.anchor = GridBagConstraints.EAST;
        c.insets = new Insets(20, 10, 12, 10);
        grid.setConstraints(buttons, c);
        add(buttons);

	// Pack and show:
	reset();
	pack();
	GUI.center(this);
	setResizable(false);
	show();
    }

    private Choice addChoice(final String label) {

        final Label newLabel = makeLabel(label);
	c.gridx = 0; c.gridy++; c.gridwidth = 1;
        c.anchor = GridBagConstraints.EAST;
        if (bFirstChoice) c.insets = new Insets(20, 12, 5, 0);
	else c.insets = new Insets(0, 12, 5, 0);
        grid.setConstraints(newLabel, c);
        add(newLabel);
        final Choice newChoice = new Choice();
        c.gridx = 1; c.gridwidth = 2;
        c.anchor = GridBagConstraints.WEST;
        if (bFirstChoice) c.insets = new Insets(20, 0, 5, 13);
	else c.insets = new Insets(0, 0, 5, 13);
        grid.setConstraints(newChoice, c);
        add(newChoice);

	bFirstChoice = false;
	return newChoice;
    }

    void reset() {

	final int nrtypes = NJ.types.length;
	final int type = bFirstReset ? nrtypes : typeChoice.getSelectedIndex();
	typeChoice.removeAll();
        for (int i=0; i<nrtypes; ++i) typeChoice.addItem(NJ.types[i]);
	typeChoice.addItem("All");
	typeChoice.select(type);

	final int nrclusters = NJ.clusters.length;
	final int cluster = bFirstReset ? nrclusters : clusterChoice.getSelectedIndex();
	clusterChoice.removeAll();
        for (int i=0; i<nrclusters; ++i) clusterChoice.addItem(NJ.clusters[i]);
	clusterChoice.addItem("All");
	clusterChoice.select(cluster);

	bFirstReset = false;
    }

    private Label makeLabel(String label) {
	if (IJ.isMacintosh()) label += "  ";
	return new Label(label);
    }
    
    public void actionPerformed(final ActionEvent e) {

	if (e.getSource() == runButton) {
	    final Tracings tracings = NJ.nhd.tracings();
	    final int nrtracings = tracings.nrtracings();
	    final int type = typeChoice.getSelectedIndex();
	    final int cluster = clusterChoice.getSelectedIndex();

	    if (statistCheckbox.getState() == true) {
		int iCount = 0;
		final double[] lengths = new double[nrtracings];
		for (int n=0; n<nrtracings; ++n) {
		    final Tracing tracing = tracings.element(n);
		    if ((tracing.type() == type || type == NJ.types.length) &&
			(tracing.cluster() == cluster || cluster == NJ.clusters.length))
			lengths[iCount++] = tracing.length();
		}
		final StringBuffer measures = new StringBuffer();
		measures.append(NJ.imagename);
		measures.append("\t");
		measures.append(cluster==NJ.clusters.length ? "All" : NJ.clusters[cluster]);
		measures.append("\t");
		measures.append(type==NJ.types.length ? "All" : NJ.types[type]);
		measures.append("\t");
		if (iCount > 0) {
		    double dSum = lengths[0];
		    double dMin = dSum;
		    double dMax = dMin;
		    for (int i=1; i<iCount; ++i) {
			final double dLen = lengths[i];
			dSum += dLen;
			if (dLen < dMin) dMin = dLen;
			else if (dLen > dMax) dMax = dLen;
		    }
		    final double dMean = dSum/iCount;
		    double dSumDev2 = 0.0;
		    for (int i=0; i<iCount; ++i) {
			final double dDev = lengths[i] - dMean;
			dSumDev2 += dDev*dDev;
		    }
		    final double dStDev = Math.sqrt(dSumDev2/(iCount-1));
		    measures.append(String.valueOf(iCount));
		    measures.append("\t");
		    measures.append(d2s(dSum));
		    measures.append("\t");
		    measures.append(d2s(dMean));
		    measures.append("\t");
		    measures.append(d2s(dStDev));
		    measures.append("\t");
		    measures.append(d2s(dMin));
		    measures.append("\t");
		    measures.append(d2s(dMax));
		} else {
		    measures.append("0");
		}
		if (NJ.srw == null || !NJ.srw.isShowing()) {
		    NJ.write("Writing measurements to new statistics results window");
		    NJ.srw = new TextWindow("NeuronJ: Statistics",statistHeadings,measures.toString(),600,300);
		} else {
		    NJ.write("Writing measurements to statistics results window");
		    final TextPanel tp = NJ.srw.getTextPanel();
		    if (clearCheckbox.getState() == true) tp.setColumnHeadings(statistHeadings);
		    tp.append(measures.toString());
		}

	    }
	    if (individCheckbox.getState() == true) {
		final StringBuffer measures = new StringBuffer();
		int iCount = 0;
		for (int n=0; n<nrtracings; ++n) {
		    final Tracing tracing = tracings.element(n);
		    if ((tracing.type() == type || type == NJ.types.length) &&
			(tracing.cluster() == cluster || cluster == NJ.clusters.length)) {
			++iCount;
			measures.append(NJ.imagename);
			measures.append("\t");
			measures.append("N"+tracing.id());
			measures.append("\t");
			measures.append(NJ.clusters[tracing.cluster()]);
			measures.append("\t");
			measures.append(NJ.types[tracing.type()]);
			measures.append("\t");
			measures.append(tracing.label());
			measures.append("\t");
			measures.append(d2s(tracing.length()));
			measures.append("\n");
		    }
		}
		if (iCount == 0) {
		    measures.append(NJ.imagename);
		    measures.append("\t");
		    measures.append("None");
		    measures.append("\t");
		    measures.append(cluster==NJ.clusters.length ? "All" : NJ.clusters[cluster]);
		    measures.append("\t");
		    measures.append(type==NJ.types.length ? "All" : NJ.types[type]);
		}
		if (NJ.trw == null || !NJ.trw.isShowing()) {
		    NJ.write("Writing measurements to new individual results window");
		    NJ.trw = new TextWindow("NeuronJ: Tracings",individHeadings,measures.toString(),600,300);
		} else {
		    NJ.write("Writing measurements to individual results window");
		    final TextPanel tp = NJ.trw.getTextPanel();
		    if (clearCheckbox.getState() == true) tp.setColumnHeadings(individHeadings);
		    tp.append(measures.toString());
		}
	    }
	} else if (e.getSource() == closeButton) { quit(); }
    }
    
    void quit() {
	setVisible(false);
	dispose();
	NJ.closeMeasures();
    }

    private String d2s(final double d) {
	if (Double.isNaN(d) || Double.isInfinite(d)) return Double.toString(d);
	else return ((FMath.round(d)==d) ? IJ.d2s(d,0) : IJ.d2s(d,3));
    }

}
